import type { AnyStatement } from 'ts-common/src/model/query/logic/general';
import type { QueryCacheData } from './model';
import { log } from 'ts-common/src/logger/logger';

export function createWhereLogic(op: string, left: string, whereLogic: string, cacheData: QueryCacheData): [string, string] {
  const newWhereLogic = `${left.replace('.', '_')}_${op}`;
  if (whereLogic) {
    whereLogic += ', ';
  }
  const remainingNodes: string[] = [];

  for (const entity of cacheData.entities) {
    if (entity.id !== left) {
      remainingNodes.push(entity.id);
    }
  }

  // TODO: Relation temporarily ignored due to unnecessary added complexity in the query
  // for (const relation of cacheData.relations) {
  //     if (relation.id !== left) {
  //         remainingNodes.push(relation.id);
  //     }
  // }

  const remainingNodesStr = remainingNodes.length > 0 ? `, ${remainingNodes.join(', ')}` : '';
  const newWithLogic = `${whereLogic}${op}(${left}) AS ${newWhereLogic}${remainingNodesStr}`;

  return [newWhereLogic, newWithLogic];
}

export function extractLogicCypher(logicQuery: AnyStatement, cacheData: QueryCacheData): [string, string] {
  switch (typeof logicQuery) {
    case 'object':
      if (Array.isArray(logicQuery)) {
        let op = logicQuery[0].replace('_', '').toLowerCase();
        const right = logicQuery?.[2];
        const [left, whereLogic] = extractLogicCypher(logicQuery[1], cacheData);
        let whereLogicMutable = whereLogic;

        switch (op) {
          case '!=':
            op = '<>';
            break;
          case '==':
            if (right === '".*"') op = '=~'; // in case "equals to <empty string>", this is what happens
            else op = '=';
            break;
          case 'like':
            op = '=~';
            break;
          case 'isempty':
            return [`(${left} IS NULL OR ${left} = "")`, whereLogicMutable];
          case 'lower':
            return [`toLower(${left})`, whereLogicMutable];
          case 'upper':
            return [`toUpper(${left})`, whereLogicMutable];
          case 'avg':
          case 'count':
          case 'max':
          case 'min':
          case 'sum':
            return createWhereLogic(op, left, whereLogicMutable, cacheData);
        }
        if (logicQuery.length > 2 && logicQuery[2]) {
          const [right, whereLogicRight] = extractLogicCypher(logicQuery[2], cacheData);
          let rightMutable = right;

          if (whereLogicRight) {
            if (whereLogicMutable) {
              whereLogicMutable += ', ';
            }
            whereLogicMutable += whereLogicRight;
          }
          if (op === '=~') {
            rightMutable = `(".*" + ${rightMutable} + ".*")`;
          }
          return [`(${left} ${op} ${rightMutable})`, whereLogicMutable];
        }
        return [`(${op} ${left})`, whereLogicMutable];
      }
      return [logicQuery, ''];
    case 'string':
      return [logicQuery.replace('@', ''), ''];
    case 'number':
      return [logicQuery.toString(), ''];
    default:
      return [logicQuery as any, ''];
  }
}
